#include <nSystem.h>
#include "torneo.h"

/*
 * Maximo paralelismo es la maxima cantidad de enfrentamientos
 * que se realizaron en un instante dado.  Debe coincidir
 * con la cantidad de mesas disponible.
 *
 * Paralelismo acumulado: mientras mayor mejor!
 *
 * Tiempo de referencia: cuando demoro con mi solucion en mi computador.
 * La medicion del tiempo no es muy precisa por lo que podria variar
 * de computador a computador.  Me gustaria ver las cifras
 * que obtienen en uch.ing.cursos.cc41b.
 */

static int mesas[4];
static int jugadores[8];
static int jugados[8][8];
static nMonitor mutex;
static int cont, en_curso, max_par, suma;
static int paralelo= TRUE;
static int msgs= TRUE;

Par partidos[4*7]= {
    {0, 4}, {1, 5}, {2, 6}, {3, 7},
    {0, 1}, {2, 4}, {3, 5}, {7, 6},
    {0, 2}, {3, 1}, {7, 4}, {6, 5},
    {0, 3}, {7, 2}, {6, 1}, {5, 4},
    {0, 7}, {6, 3}, {5, 2}, {4, 1},
    {0, 6}, {5, 7}, {4, 3}, {1, 2},
    {0, 5}, {4, 6}, {1, 7}, {2, 3} };

static int referencia[5]= {0, 1280, 640, 500, 450};
static int tiempo_inicio[4*7];

typedef int Duraciones[8];


static int duraciones[8][8]= {
  /*       0   1   2   3   4   5   6   7  */
  /* 0 */  0,  10,  50,  20,  50,  60,  100, 30,
  /* 1 */  0,   0,  40,  70,  60,  30,  10,  50,
  /* 2 */  0,   0,   0,  20,  70,  80,  20,  90,
  /* 3 */  0,   0,   0,   0,  40,  30,  60,  50,
  /* 4 */  0,   0,   0,   0,   0,  70,  10,  40,
  /* 5 */  0,   0,   0,   0,   0,   0,  90,  10,
  /* 6 */  0,   0,   0,   0,   0,   0,   0,  20,
  /* 7 */  0,   0,   0,   0,   0,   0,   0,   0
};

static void ini_torneo(int N) {
  int i, j, m;
  cont= en_curso= max_par= suma= 0;
  for (j= 0; j<8; j++) {
    jugadores[j]= FALSE;
    for (i=0; i<8; i++)
      jugados[i][j]= FALSE;
  }
  for (m= 0; m<N; m++)
    mesas[m]= FALSE;
}
 
void valida_secuencia();

int nMain(int argc, int argv) {
  int N, tiempo_inicial;
  mutex= nMakeMonitor(1);
  msgs= TRUE;
  for (N= 1; N<=4; N++) {
    nPrintf("\n-------------------------------------------------\n");
    nPrintf("Torneo con %d mesas\n\n", N);
    ini_torneo(N);
    tiempo_inicial= nGetTime();

    torneo(N);
   
    valida_secuencia();
    if (cont!=28)
      nFatalError("nMain", "no se jugaron todos los partidos para N=%d\n",
                  N);
    nPrintf("Resultados para N=%d\n", N);
    nPrintf("maximo paralelismo: %d\n", max_par);
    if (max_par!=N) {
      nPrintf("*** advertencia *** el paralelismo es inferior al esperado\n");
      paralelo= FALSE;
    }
    nPrintf("paralelismo acumulado: %d\n", suma);
    nPrintf("tiempo transcurrido: %d milisegs\n", nGetTime()-tiempo_inicial);
    nPrintf("tiempo de referencia: %d milisegs\n", referencia[N]);
  }

  nPrintf("Test de robustez en modo non preemptive durante 10 segundos.\n");
  msgs= FALSE;
  nSetTimeSlice(1);
  {
    int tiempo_actual= tiempo_inicial= nGetTime();
    int tiempo_msg= tiempo_actual+1000;
    int iter= 1;
    tiempo_inicial= tiempo_actual;
    while (tiempo_actual-tiempo_inicial<10000) {
      if (tiempo_actual>tiempo_msg) {
        nPrintf("%d ", iter++);
        tiempo_msg= tiempo_inicial+1000*iter;
      }
      for (N= 1; N<=4; N++) {
        ini_torneo(N);
        torneo(N);
        if (cont!=28)
          nFatalError("nMain", "no se jugaron todos los partidos para N=%d\n",
                      N);
      }
      tiempo_actual= nGetTime();
    }
  }
  nPrintf("10\nTest de robustez finalizado con exito.\n");

  if (!paralelo) {
    nPrintf("*** Advertencia ***:\n%s\n%s\n%s\n%s\n",
    "En uno de los tests no se alcanzo el paralelismo esperado.",
    "Intente correr los tests en un computador sin carga adicional.",
    "Si el problema persiste, revise su tarea, pues no cumple con el",
    "requisito de paralelismo.");
  }
  return 0;
}

void enfrentar(int i, int j, int m) {
  int posicion;
  nEnter(mutex);
  if (i<0 || i>=8)
    nFatalError("enfrentar", "numero de jugador %d fuera de rango\n", i);
  if (j<0 || j>=8)
    nFatalError("enfrentar", "numero de jugador %d fuera de rango\n", j);
  if (i==j)
    nFatalError("enfrentar", "jugador %d se enfrenta consigo mismo\n", i);
  if (jugadores[i])
    nFatalError("enfrentar", "jugador %d ya esta jugando\n", i);
  if (jugadores[j])
    nFatalError("enfrentar", "jugador %d ya esta jugando\n", j);
  if (mesas[m])
    nFatalError("enfrentar", "mesa %d ya esta ocupada\n", m);
  if (jugados[i][j])
    nFatalError("enfrentar", "partido entre %d y %d ya se jugo\n", i, j);
  /* Anotar tiempo de inicio */
  {
    int ronda_prev;
    int ctime= nGetTime();
    for (posicion= 0; posicion<28; posicion++)
      if (partidos[posicion].x==i && partidos[posicion].y==j ||
          partidos[posicion].x==j && partidos[posicion].y==i)
        break;
    if (posicion>=28)
      nFatalError("enfrentar", "no se encontro partido (%d,%d)\n", i, j);
    tiempo_inicio[posicion]= ctime;
  }

  if (msgs)
    nPrintf("enfrentando %d con %d en mesa %d\n", i, j, m);
  jugadores[i]= jugadores[j]= mesas[m]= TRUE;
  jugados[i][j]= jugados[j][i]= TRUE;
  en_curso++;
  if (en_curso>max_par)
    max_par= en_curso;
  cont++;
  suma+= en_curso;
  nExit(mutex);
  if (msgs) {
    if (i<j)
      nSleep(duraciones[i][j]);
    else
      nSleep(duraciones[j][i]);
  }
  else
    nSleep(0);
  nEnter(mutex);
  jugadores[i]= jugadores[j]= mesas[m]= FALSE;
  if (msgs)
    nPrintf("finaliza partido entre %d y %d en mesa %d\n", i, j, m);
  en_curso--;
  nNotifyAll(mutex);
  nExit(mutex);
}

void valida_secuencia() {
  int ronda, k;
  int tiempo_ult= tiempo_inicio[0];
  /* validar primera ronda */
  for (k= 1; k<4; k++) {
    if (tiempo_inicio[k]-tiempo_inicio[k-1]<-2)
      nFatalError("valida_secuencia", "partido %d comenzo antes que %d\n",
                  k, k-1);
    if (tiempo_inicio[k]>tiempo_ult)
      tiempo_ult= tiempo_inicio[k];
  }
  /* validar rondas restantes */
  for (ronda= 1; ronda<4; ronda++) {
    int prev_ult= tiempo_ult;
    int prox_ronda;

    k= ronda*4;
    prox_ronda= k+4;
    tiempo_ult= 0;
    while (k<prox_ronda) {
      if (tiempo_inicio[k]-prev_ult<-2)
        nFatalError("valida_secuencia", "partido %d comenzo antes que "
          "uno de la ronda previa\n", k);
      if (tiempo_inicio[k]>tiempo_ult)
        tiempo_ult= tiempo_inicio[k];
      k++;
    }
  }
}
