#include <nSystem.h>
#include "torneo.h"

/*
 * Maximo paralelismo es la maxima cantidad de enfrentamientos
 * que se realizaron en un instante dado.  Debe coincidir
 * con la cantidad de mesas disponible.
 *
 * Paralelismo acumulado: mientras mayor mejor!
 *
 * Tiempo de referencia: cuando demoro con mi solucion en mi computador.
 * La medicion del tiempo no es muy precisa por lo que podria variar
 * de computador a computador.  Me gustaria ver las cifras
 * que obtienen en uch.ing.cursos.cc41b.
 */

static int mesas[4];
static int jugadores[8];
static int jugados[8][8];
static nSem mutex;
static int cont, en_curso, max_par, suma;

Par partidos[4*7]= {
    {0, 4}, {1, 5}, {2, 6}, {3, 7},
    {0, 1}, {2, 4}, {3, 5}, {7, 6},
    {0, 2}, {3, 1}, {7, 4}, {6, 5},
    {0, 3}, {7, 2}, {6, 1}, {5, 4},
    {0, 7}, {6, 3}, {5, 2}, {4, 1},
    {0, 6}, {5, 7}, {4, 3}, {1, 2},
    {0, 5}, {4, 6}, {1, 7}, {2, 3} };

static int referencia[5]= {0, 1280, 640, 500, 450};

int nMain(int argc, int argv) {
  int N, tiempo_inicial;
  mutex= nMakeSem(1);
  for (N= 1; N<=4; N++) {
    nPrintf("\n-------------------------------------------------\n");
    nPrintf("Torneo con %d mesas\n\n", N);
    int i, j, m;
    cont= en_curso= max_par= suma= 0;
    for (j= 0; j<8; j++) {
      jugadores[j]= FALSE;
      for (i=0; i<8; i++)
        jugados[i][j]= FALSE;
    }
    for (m= 0; m<N; m++)
      mesas[m]= FALSE;

    tiempo_inicial= nGetTime();
    torneo(N);
   
    if (cont!=28)
      nFatalError("nMain", "no se jugaron todos los partidos para N=%d\n",
                  N);
    nPrintf("Resultados para N=%d\n", N);
    nPrintf("maximo paralelismo: %d\n", max_par);
    nPrintf("paralelismo acumulado: %d\n", suma);
    nPrintf("tiempo transcurrido: %d milisegs\n", nGetTime()-tiempo_inicial);
    nPrintf("tiempo de referencia: %d milisegs\n", referencia[N]);
  }

  return 0;
}

static int duracion[8][8]= {
  /*       0   1   2   3   4   5   6   7  */
  /* 0 */  0,  10,  50,  20,  50,  60,  100, 30,
  /* 1 */  0,   0,  40,  70,  60,  30,  10,  50,
  /* 2 */  0,   0,   0,  20,  70,  80,  20,  90,
  /* 3 */  0,   0,   0,   0,  40,  30,  60,  50,
  /* 4 */  0,   0,   0,   0,   0,  70,  10,  40,
  /* 5 */  0,   0,   0,   0,   0,   0,  90,  10,
  /* 6 */  0,   0,   0,   0,   0,   0,   0,  20,
  /* 7 */  0,   0,   0,   0,   0,   0,   0,   0
};
 
void enfrentar(int i, int j, int m) {
  nWaitSem(mutex);
  if (i==j)
    nFatalError("enfrentar", "jugador %d se enfrenta consigo mismo\n", i);
  if (jugadores[i])
    nFatalError("enfrentar", "jugador %d ya esta jugando\n", i);
  if (jugadores[j])
    nFatalError("enfrentar", "jugador %d ya esta jugando\n", j);
  if (mesas[m])
    nFatalError("enfrentar", "mesa %d ya esta ocupada\n", m);
  if (jugados[i][j])
    nFatalError("enfrentar", "partido entre %d y %d ya se jugo\n", i, j);
  nPrintf("enfrentando %d con %d en mesa %d\n", i, j, m);
  jugadores[i]= jugadores[j]= mesas[m]= TRUE;
  jugados[i][j]= jugados[j][i]= TRUE;
  en_curso++;
  if (en_curso>max_par)
    max_par= en_curso;
  cont++;
  suma+= en_curso;
  nSignalSem(mutex);
  if (i<j)
    nSleep(duracion[i][j]);
  else
    nSleep(duracion[j][i]);
  nWaitSem(mutex);
  jugadores[i]= jugadores[j]= mesas[m]= FALSE;
  nPrintf("finaliza partido entre %d y %d en mesa %d\n", i, j, m);
  en_curso--;
  nSignalSem(mutex);
}
